;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; File Name: Kernel.s
;; Function : Implements Large Memory Model kernel for ImageCraft
;;            compiler.
;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Kernel primitives
; All kernel primitives lies in the range of 0x18 to 0xBF COG RAM.

	.area text(rom,rel)
__Kernel::
;;;;;;;; JUMP TABLE 0x0 to 0x14 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	jmp #finit	;finit
	jmp #fcall	;fcall
	jmp #ficall	;ficall - indirection call
	jmp #fret	;fret
	jmp #fload	;fload
	jmp #fenter	;fenter
	jmp #fpop	;fpop
	jmp #fldcstk	;fetch from the cog stack pointed by CRSP
	jmp #fstcstk	;fstore to cog stack pointed by CRSP
	jmp #fbytecopy	;byte copy loop
	jmp #fwordcopy	;word copy loop
	jmp #flongcopy	;long copy loop
	jmp #fdiv32s	;signed division
	jmp #fdiv32u	;unsigned division
	jmp #fmul32	;multiplication
	.long __start	;PC   - Program Counter         (0xF)
	.long 0x8000	;HRSP - HUB RAM stack pointer   (0x10)
	.long 0		;CRSP - COG RAM stack pointer   (0x11)
	.long 0		;fcache - future support	(0x12)
	.long 0		;fsystem - future support	(0x13)
	.long 0		;debug - future support		(0x14)
;;;;;;;;;;; end of jump table ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	
;;;;;;;;;;;; Kernel Primitives 0x15 to BF ;;;;;;;;;;;;;;;;;;;;;;;
;fcache - When the "kernel" executes an "FCACHE" instruction, which in reality just calls a small
;	  primitive routine in the cog, it will copy all longs after the FCACHE to the cog's execution
;	  buffer stopping only when it runs across a "NULL" long (0).
; To be implemented in future

;fret - loads PC from word at SP, decrements SP by two.
;fret - 0x15

fret:
	movs frt, CRSP
	add     CRSP,#1
frt:	mov     PC,0-0
        jmp     #next

;POPs the registers from the stack - Top of the stack pointed by CRSP
;fpop - 0x19
fpop:
	rdlong RegList,PC
	add PC,#4
	cmp RegList,#0	WZ
IF_E	jmp #next	
	movs read,CRSP
	movd read,RegList
	;nop
	add CRSP,#1
read:	mov 0-0,0-0
	jmp #fpop

;fcall - Increments SP by 2, replaces PC with long at PC after it saves PC+4 at SP.
;fcall - 0x23
fcall:
	rdlong ftemp,PC
	sub    CRSP,#1
	movd S1, CRSP
    ;nop
	add PC,#4	;fill pipe
S1:	mov  0-0,PC
	mov    PC,ftemp
        jmp #next

;ficall - 0x2B
ficall:
	sub    CRSP,#1
	movd icall, CRSP
	nop
icall:	mov  0-0,PC
	mov    PC,TEMP0
        jmp #next

;fload - This primitive copy the long following the instruction to a specific
;location (then skip the 2nd long too). This allows a long (often a HUB address) to be loaded
;without a complex multiple instruction sequence.
;fload - 0x31

fload:  rdlong  TEMP0,PC    ; get the data
        add     PC,#4
        jmp     #next

;next - This will fetch one instruction at a time from HUB RAM to COG RAM
;and executes it. To reduce the execution speed the loop is unrolled into 4.
;PC should point to 1st instruction in HUB memory.
;next - 0x35

next:   rdlong  inst1,PC
        add     PC,#4
inst1:  nop
        rdlong  inst2,PC
        add     PC,#4
inst2:  nop
        rdlong  inst3,PC
        add     PC,#4
inst3:  nop
        rdlong  inst4,PC
        add     PC,#4
inst4:  nop        
        rdlong  inst5,PC
        add     PC,#4
inst5:  nop
        rdlong  inst6,PC
        add     PC,#4
inst6:  nop
        rdlong  inst7,PC
        add     PC,#4
inst7:  nop
        rdlong  inst8,PC
        add     PC,#4
inst8:  nop

; Check the enable command in USER5. If non-zero enable interrupts.
; We cant write USERn in our asm.s files, and enabling USER0 in C code causes crash.
; Let C code signal the operation by setting USER 5
/*        
        tjz     USER5,  #noenab
        mov     USER4,  #0
        wrlong  USER4,  USER3   ; clear isr
        mov     USER0,  #1        
        mov     USER5,  #0
noenab:
*/
        //jmp     #next           ; debug

; read hw status here to stay in hub cycle

        rdlong  USER4,  USER3   ; read hw isr

        ;mov     tmp2,   USER0   ; we need to rdlong this ?
        ;rdlong  temp,   USER0
        ;tjz     temp,   #next   ; USER0 is interrupt enable. if clear fetch next
        tjz     USER0,  #next   ; USER0 is interrupt enable. if clear fetch next

        tjnz    USER4,  #doint  ; USER4 is isr content. if non-zero, do interrupt

        jmp     #next           ; if no interrupts, fetch next instruction

; First level callback function address stored in USER2 should only call
; another function to preserve context/stack integrity (no variables).
; The function called by first level callback can have variables, etc...

        ;tjz     USER1,  #next   ; USER1 is user interrupt vector

doint:  sub     CRSP,   #1
        movd    callbk, CRSP
        mov     USER0,  #0      ; must disable interrupts for service in kernel
callbk: mov     0-0,    PC      ; Save only the PC. 
    	mov     PC,     USER2   ; USER2 contains first level callback address

        ;or  dira, #3
        ;xor outa, #3

        jmp     #next

        
;finit - Initialises PC to start up routine and calls 'next' kernel primitive.        
;finit - 0x42

finit:
	mov TEMP0,PAR
	rdlong PC,TEMP0
	add TEMP0,#4
	rdlong HRSP,TEMP0
 	mov CRSP, #$18F
 	jmp #next


;fenter - Create a stack frame and push registers
;fenter - 0x48

fenter:
	rdlong RegList,PC	WZ
	add PC,#4
IF_E	jmp #next	
	sub CRSP,#1
	movd reglst,CRSP
	movs reglst,RegList
	nop
reglst: mov 0-0,0-0
	jmp #fenter

;fldcstk - Fetch the local variables from stack
;fldcstk - 0x52
fldcstk: 
	mov stack1, CRSP
	add stack1, TEMP0
	movs fetch1, stack1
	nop
fetch1: mov TEMP0, 0-0
	jmp #next
	
;fstcstk - store the local variables to stack
;fstcstk - 0x58
fstcstk:
	mov stack2, CRSP
	add stack2, TEMP1
	movd store, stack2
	nop
store:	mov 0-0, TEMP0
	jmp #next
	
;fbytecopy - 0x5E
fbytecopy:
	rdbyte TEMP0,R1
	wrbyte TEMP0,R0
	add R0,#1
	add R1,#1
	sub R2,#1	WZ
IF_NZ	jmp #fbytecopy
	add R2,#1
	jmp #next	

;fwordcopy - 0x66
fwordcopy:
	rdword TEMP0,R1
	wrword TEMP0,R0
	add R0,#2
	add R1,#2
	sub R2,#2	WZ
IF_NZ	jmp #fwordcopy
	add R2,#2
	jmp #next

;flongcopy - 0x6E
flongcopy:
	rdlong TEMP0,R1
	wrlong TEMP0,R0
	add R0,#4
	add R1,#4
	sub R2,#4	WZ
IF_NZ	jmp #flongcopy
	add R2,#4
	jmp #next


;fdiv32s - 0x76	
;fdiv32s - Signed 32 bit division
        ;Divisor --> R1
        ;Dividend--> R0
        ;Result --> Quotient in R0
        ;           Remainder in R1

fdiv32s:
        mov temp,#32
        mov tmp2, #0		; tmp2 is temp register to hold the remainder
	mov ftemp, R1	
	shr ftemp, #31  WZ	; get the divisor sign (ftemp)
IF_NZ	sub tmp2, R1		; divisor = -divisor if < 0
IF_NZ	mov R1, tmp2
	mov ftemp2, R0	
	shr ftemp2, #31  WZ	; get the dividend sign (ftemp2)
	mov tmp2, #0
IF_NZ	sub tmp2, R0		; dividend = -dividend if < 0
IF_NZ	mov R0, tmp2
	mov tmp2, #0
	xor ftemp, ftemp2	; sign of result (ftemp)
up:
	shl R0,#1    WC
        rcl tmp2,#1    WC
        cmp R1,tmp2    WC WZ
IF_A	jmp #down1		;Branch to down1
	sub tmp2,R1
        add R0,#1
down1:
	sub temp, #1	WZ
IF_NE	jmp #up			;Branch to up
        mov R1,tmp2
        mov tmp2, #0
        cmp ftemp2, #0   WZ
IF_NZ	sub tmp2, R1		; adjust the sign of remainder
IF_NZ	mov R1, tmp2
	mov tmp2, #0
	cmp ftemp, #0   WZ	
IF_NZ	sub tmp2, R0
IF_NZ	mov R0, tmp2		; adjust the sign of quotient
	jmp #next

;fdiv32u - 0x96	
;fdiv32u - Unsigned 32 bit division
        ;Divisor --> R1
        ;Dividend--> R0
        ;Result --> Quotient in R0
        ;           Remainder in R1

fdiv32u:
        mov ftemp,#32
	mov ftemp2, #0
up2:
	shl r0,#1       WC
        rcl ftemp2,#1       WC
        cmp R1,ftemp2 	WC WZ
IF_A	jmp #down	;Branch to down
	sub ftemp2,R1
        add R0,#1
down:
	sub ftemp, #1	WZ
IF_NE	jmp #up2	;Branch to up
        mov R1,ftemp2
	jmp #next	


;fmul32 - 0xA3
;fmul32	- multiplication
	;R0 --> 1st operand (32 bit)
	;R1 --> 2nd operand (32 bit)
	;Result --> R0 (<= 32 bit)

fmul32:
	mov ftemp,#0
start:
	cmp R0,#0	WZ
IF_E	jmp #down3	;Branch to down3
	shr R0,#1	WC
IF_AE	jmp #down2	;Branch to down2
	add ftemp,R1	WC
down2:
	shl R1,#1	WC
	jmp #start	;Branch to start
down3:
	mov R0,ftemp
	jmp #next
	
RegList:	.long 0			;(0xAD)
reg:		.long 0			;(0xAE)
count:		.long 0			;(0xAF)
temp:		.long 0			;(0xB0)
tmp2:		.long 0			;(0xB1)
mask:		.long 0x0000ffff	;(0xB2)
ftemp:		.long	0		;(0xB3)
ftemp2:		.long	0		;(0xB4)
stack1:		.long 0			;(0xB5)
stack2:		.long 0			;(0xB6)
		
		res 334

;fsystem - To be implemented
;debug - To be implemented

;Kernel primitives should not exceed 0xBF
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;0xC0 to 0xF9 - FCACHE code buffer - future use
;;;; 0xFA to 0xFF - TEMPx registers;;;;;;;;;;;;;;;;;;;;;;
;0xFA - TEMP0
;0xFB - TEMP1
;0xFC - TEMP2
;0xFD - TEMP3
;0xFE - TEMP4
;0xFF - TEMP5
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;0x100 to 0x10F - Virtual registers from R0 to R15;;;;
;0x100 - R0
;0x101 - R1
;0x102 - R2
;0x103 - R3
;0x104 - R4
;0x105 - R5
;0x106 - R6
;0x107 - R7
;0x108 - R8
;0x109 - R9
;0x10A - R10
;0x10B - R11
;0x10C - R12
;0x10D - R13
;0x10E - R14
;0x10F - R15
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;0x110 to 0x18F - FCALL and data stack, stack grows downward
;CRSP initialized to 18F in finit kernel routine

;0x190 to 0x1EF - Debug and multiprocessing support - future support
;0x1EA - USER0
;0x1EB - USER1
;0x1EC - USER2
;0x1ED - USER3
;0x1EE - USER4
;0x1EF - USER5
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;0x1F0 to 0x1FF - Parallax special registers;;;;;;;;;
;0x1F0 - DIRA
;0x1F1 - DIRB
;0x1F2 - INA
;0x1F3 - INB
;0x1F4 - OUTA
;0x1F5 - OUTB
;0x1F6 - CNT
;0x1F7 - CTRA
;0x1F8 - CTRB
;0x1F9 - FRQA
;0x1FA - FRQB
;0x1FB - PHSA
;0x1FC - PHSB
;0x1FD - VCFG
;0x1FE - VSCL
;0x1FF - PAR
;;;;;;;;;;;;;;;;; END OF KERNEL ;;;;;;;;;;;;;;;;;;;;;;;;;
