/**
 * @file interrupt.c
 * Defines ICCPROP kernel interrupt interface
 */
#include "interrupt.h"
#include "IntArray.h"

/*
 * This is the first level interrupt handler.
 * It calls the user handler registered during init after disabling interrupts.
 * It re-enables global interrupts before function return.
 * Adding variables to this function will break the system.
 */
void int_interrupt(void)
{
    user_handler();
}

/*
 * Initialize interrupt subsystem to known disabled state.
 * Set first level handler in kernel, and set user handler.
 * This function will not enable interrupts.
 * See header for more details.
 */
void int_initialize(hwint_t* hwp)
{
    int usr2 = (int)&int_interrupt; // first level interrupt handler
    int usr3 = (int)&hwp->isr;      // hw isr address pointer
    asm("mov USER0, #0");           // disable interrupts
    asm("mov USER1, #0");           // clear "vector"
    asm("mov USER2, %usr2");        // set first level interrupt handler
    asm("mov USER3, %usr3");        // set isr address pointer
    asm("mov USER4, #0");           // clear isr contents register
    // let user enable interrupts
}

/*
 * delay function used after start cog and during screen updates
 * @param per is number of ticks to wait ....
 * Careful, if per is too small, the processor will wait until the next count
 * which can be on the order of minutes.
 */
void int_pause(int per)
{
    int time = 0;
    asm("mov        %time, cnt");
    asm("add        %time, %per");
    asm("waitcnt    %time, %per");
}

/*
 * Start the hardware interrupt monitor COG.
 * See header for more details.
 */
int int_start(hwint_t* hwp)
{
    int id = 0;
    int_initialize(hwp);
    id = cognewNative(IntArray, hwp);
    int_pause(10000000); // wait for driver to start
    return id;
}

/*
 * Stop hardware interrupt monitor cog
 * See header file for more details.
 */
void int_stop(int id)
{
    if(id) {
        asm("cogstop %id");
    }
}

/*
 * Gets the vector. Called at interrupt time.
 * See header for more details.
 */
size_t int_getVector(void)
{
    size_t rc;
    asm("mov %rc, USER1");
    return rc;
}

/*
 * Sets the vector. Called at interrupt time.
 * See header for more details.
 */
void int_setVector(size_t handler)
{
    size_t rc = handler;
    asm("mov USER1,%rc");
}

/*
 * enable all interrupts
 */
void enable(void)
{
    asm("mov USER0, #1");
}

/*
 * disable all interrupts
 */
void disable(void)
{
    asm("mov USER0, #0");
}
