/**
 * @file interrupt.h
 * Defines ICCPROP kernel interrupt interface API
 */
#ifndef __INTERRUPT__
#define __INTERRUPT__

#include <stdlib.h>

/**
 * User's interrupt handler. 
 */
void user_handler(void);

/**
 * defines hardware monitor COG interrupt interface
 */
typedef struct hwint_struct
{
    long isr;   // status register          change without restart
    long imr;   // mask register            change without restart
    long pins;  // pin bit-map register     restart device to change this
    long invt;  // invert interrupt sense   restart device to change this
} hwint_t;
 
/**
 * This is the first level interrupt handler.
 * It calls the user handler registered during init after disabling interrupts.
 * It re-enables global interrupts before function return.
 * Adding variables to this function will break the system.
 */
void int_interrupt(void);

/**
 * Initialize interrupt subsystem to known disabled state.
 * Set first level handler in kernel, and set user handler.
 * This function will not enable interrupts.
 * @param hwp is pointer to pseudo-hardware interrupt device registers.
 */
void int_initialize(hwint_t* hwp);

/**
 * Start the hardware interrupt monitor COG.
 * @param hwp is pointer to the COG hwint_t interface.
 * @returns new cogid
 */
int  int_start(hwint_t* hwp);

/**
 * Stop the hardware interrupt monitor COG.
 * @param id is COG id.
 */
void int_stop(int id);

/**
 * Gets the vector. Called at interrupt time.
 * If the vector is -1, a hardware interrupt is active.
 * If the vector is > 0, a software interrupt is active.
 * If the vector = 0, no interrupt is active.
 * @returns value of vector.
 */
size_t int_getVector(void);

/**
 * Sets the vector. Called at interrupt time.
 * @param handler is pointer to the callback function.
 */
void int_setVector(size_t  handler);

/**
 * enable all interrupts
 */
void enable(void);

/**
 * disable all interrupts
 */
void disable(void);

#endif